/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */


#if defined(__LINUX__)
  #include <sys/types.h>
  #include <sys/time.h>
  #include <unistd.h>
#endif
#if defined(__WINDOWS__)
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "NCanUtils.h"

#if defined (__CAN__)
  unsigned gMainThreadId = 0;
  unsigned gProcessId = 0;
#endif

namespace NCanUtils
{
  
void SleepMs(int aMs)
{
#if defined(__WINDOWS__)
  Sleep(aMs);
#endif
#if defined(__LINUX__)
  usleep(1000*aMs);
#endif
}
  
void ArtTime(const struct timeval & Time, int & aDays, int & aMs)
{
  aMs = 1000*(Time.tv_sec % 86400) + Time.tv_usec/1000;
  aDays = (int) (Time.tv_sec / 86400);
  // 72 76 80 84 88 92 96 : 7 Schaltjahre
  aDays -= (2000-1970)*365 + (2000-1970)/4;
  if (aMs < 0)
  {
    aMs += 1000*86400;
    aDays--;
  }
}

#if defined(__LINUX__)

void ArtTime(int & aDays, int & aMs)
{
  struct timeval Time;

  if (!gettimeofday(&Time, NULL))  // seit 1. Jan. 1970
  {
    ArtTime(Time, aDays, aMs);
  } else {
    aMs = 0;
    aDays = 0;
  }
}

#endif

static const unsigned cMonthDays[13] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};
// jan feb mar  apr  mai  jun  jul  aug  sep  okt  nov  dez
//  31  28  31   30   31   30   31   31   30   31   30   31
void GetDateFromDays(int aDays, int & aDay, int & aMonth, int & aYear)
// aDays == 0 entspricht dem 1.1.2000 (aDay = 1, aMonth = 1, aYear = 2000)
{
  aMonth = 1;
  aDay = aDays;
  
  // immer 4 Jahre abzählen: 4*365 Tage + 1 Tag vom Schaltjahr
  aYear = aDay / (4*365 + 1);
  aDay -= aYear*(4*365 + 1);
  aYear = 4*aYear + 2000;
  if (aDay <= 0)
  {
    aDay += 4*365 + 1;
    aYear -= 4;
  }
  // das 1. Jahr (z.B. 2012) ist ein Schaltjahr, danach bis zu 3 normale Jahre
  if (aDay >= 366)
  {
    aYear++;
    aDay -= 366;
    
    while (aDay >= 365)
    {
      aYear++;
      aDay -= 365;
    }
  }
  aDay++; // der erste Tag im Monat ist der 1.
  while ((int) cMonthDays[aMonth] < aDay) 
  {
    if (aMonth == 2 && !(aYear & 3))
    {
      if ((int) cMonthDays[2]+1 >= aDay)  // 28. Feb. in einem Schaltjahr
        break;
      aDay--;    // für den 29. Feb. (ist nicht in cMonthDays enthalten)
    }
    aMonth++;
  }
  aDay -= cMonthDays[aMonth-1];
}

bool GetArtDays(int Year, int Month, int Day, int & Days)
{
  if (Year < 2000 || Year > 2100 ||
      Month < 1 || Month > 12 ||
      Day < 1 || Day > 31)
    return false;
  
  if (Month == 2)
  {
    if ((Year % 4) != 0 && Day >= 29)
      return false;
    
    if (Day > 29)
      return false;
  } else
  if ((unsigned) Day > cMonthDays[Month] - cMonthDays[Month-1])
    return false;
  
  Year -= 2000;
  Days = 365*Year + (Year - 1) / 4 + cMonthDays[Month-1] + Day;
  if ((Year % 4) == 0 && Month > 2)
    Days++;
  
  return true;  
}
  
void PrintDayTime(int time_ms, char * OutStr)
{
  int Sec = time_ms / 1000;
  int Min = Sec / 60;
  int Hour = Min / 60;
  
  sprintf(OutStr, "%2.2d:%2.2d:%2.2d.%3.3d",
          Hour, Min % 60, Sec % 60, time_ms % 1000);
}
  
void PrintTime(int time_days, int time_ms, char * OutStr, bool csv)
{
  int Day, Month, Year;
  
  GetDateFromDays(time_days, Day, Month, Year);
  if (csv)
  {
    sprintf(OutStr, "'%4.4d-%2.2d-%2.2d',%d", Year, Month, Day, time_ms);
  } else {
    char day_str[32];
    
    PrintDayTime(time_ms, day_str);
    sprintf(OutStr, "%d.%d.%d %s",
            Day, Month, Year, day_str);
  }
}
  
void PrintTime(const KCanFrame & Frame, char * OutStr, bool csv)
{
  PrintTime(Frame.TimeStampDay, Frame.TimeStampMs, OutStr, csv);
}  

void PrintMs(int days, int ms, char * stream)
{
  long long time_stamp = days;
  
  time_stamp += 30*365 + 30 / 4; // Tage zwischen 1970 und 2000
  time_stamp *= 24*3600*1000;
  // GMT time_stamp -= 2*3600000; 
  time_stamp += ms;
  
  // ./ms2time.sh 1399346245967  ==> 2014-05-06 05:17:25
  // date -d "2014-05-06 05:17:25" ==>  1399346245  (03:17:25 einsetzen GMT!)
  sprintf(stream, "%lld", time_stamp); 
}
  
void PrintDateTime(int days, int ms, char * stream)
{
  // format: YYYY-MM-DD HH:MM:SS.mmm
  int Day, Month, Year;
  char day_str[32];
  
  GetDateFromDays(days, Day, Month, Year);
  PrintDayTime(ms, day_str);
  
  sprintf(stream, "%4.4d-%2.2d-%2.2d %s", Year, Month, Day, day_str);
}
  
void FormatFrame(const KCanFrame & Frame, bool csv, char * OutStr)
{
  char TimeStr[64];
  
  PrintTime(Frame, TimeStr, csv);
  if (csv)
  {
    sprintf(OutStr, "%u,%s,%u,%d", Frame.Counter, TimeStr,
            Frame.Id & 0xffffffff, Frame.Flags & 0xff);
    for (int i = 0; i < 8; i++)
    {
      strcat(OutStr, ",");
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%d", Frame.Data[i]);
      else
        strcat(OutStr, "NULL");
    }
  } else {
    if (!Frame.Counter && !Frame.TimeStampDay)
      sprintf(OutStr, "%8x %2.2x [%d] ", 
              Frame.Id & 0x1fffffff, Frame.Flags & 0xff, Frame.Len);
    else
      sprintf(OutStr, "%3u  %s  %8x %2.2x [%d] ", Frame.Counter, TimeStr,
              Frame.Id & 0x1fffffff, Frame.Flags & 0xff, Frame.Len);
    for (int i = 0; i < 8; i++)
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%2.2x ", Frame.Data[i]);
      else
        sprintf(OutStr + strlen(OutStr), "   ");
    strcat(OutStr, " ");
    for (int i = 0; i < 8; i++)
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%c",
                ' ' <= Frame.Data[i] && Frame.Data[i] < 127 ? Frame.Data[i] : '.');
  }
  strcat(OutStr, "\n");
}
  
bool IsHexDigit(char aDigit)
{
  if (('0' <= aDigit && aDigit <= '9') ||
      ('a' <= aDigit && aDigit <= 'f') ||
			('A' <= aDigit && aDigit <= 'F'))
    return true;
  
  return false;
}

int GetHexDigit(char aDigit)
{
  if ('0' <= aDigit && aDigit <= '9')
    return (int) aDigit - '0';
  
  if ('A' <= aDigit && aDigit <= 'F')
    return (int) aDigit - ('A' - 10);
  
  if ('a' <= aDigit && aDigit <= 'f')
    return (int) aDigit - ('a' - 10);
  
  return -1;
}

unsigned GetHex(const char * & Line)
{
  if (!Line)
    return 0;
  
  unsigned Hex = 0;
  
  while (IsHexDigit(*Line))
  {
    Hex = 16*Hex + GetHexDigit(*Line++);
  }
  return Hex;
}
  
bool GetHexByte(const char * & Line, unsigned char & byte)
{
  if (!IsHexDigit(Line[0]) ||
      !IsHexDigit(Line[1]))
    return false;
  
  byte = 16*GetHexDigit(Line[0]) +
            GetHexDigit(Line[1]);
  Line += 2;
  
  return true;  
}
  
bool GetInt(const char * & Line, int & res)
{
  bool neg = *Line == '-';
  if (neg)
    Line++;
  bool Ok = ('0' <= *Line && *Line <= '9');
  
  res = 0;
  while ('0' <= *Line && *Line <= '9')
  {
    res = 10*res + (unsigned char)(*Line) - '0';
    Line++;
  }
  
  return Ok;
}
  
}

////////////////////////////////////////////////////////////////////////////////

KCanFrame::KCanFrame()
{
  Init();
}

void KCanFrame::Init()
{
  memset(this, 0, sizeof(*this));
}

void KCanFrame::InitElsterFrame(unsigned short sender_id,
                                unsigned short receiver_id,
                                unsigned short TelegrammId)
{
  unsigned char address = (receiver_id & 0x780) / 8;
  Init();
  Id = sender_id;
  Len = 5;
  Data[0] = (unsigned char)((address & 0xf0) + 1);
  Data[1] = (receiver_id & 7);
  Data[2] = 0xfa;
  Data[3] = (unsigned char)(TelegrammId >> 8);
  Data[4] = (unsigned char)(TelegrammId & 0xff);
}
bool KCanFrame::EqualData(const KCanFrame & Frame) const
{
  if (Len != Frame.Len)
    return false;
  
  for (int i = 0; i < Len; i++)
    if (Data[i] != Frame.Data[i])
      return false;
  
  return true;
}

bool KCanFrame::Changed(const KCanFrame & SendFrame) const
{
  if (Len < 5 || !IsAnswer(SendFrame))
    return false;
  
  return Value() == SendFrame.Value();
}

bool KCanFrame::IsAnswer(const KCanFrame & Frame) const
{
  // CAN-Ids vergleichen
  return (((Frame.Id / 8) & 0xf0) == (unsigned) (Data[0] & 0xf0)) &&
         ((Id / 8) & 0xf0) == (unsigned) (Frame.Data[0] & 0xf0);
}

bool KCanFrame::Supported() const
{
  if (Len < 5)
    return true;
  
  return Value() != 0x8000;
}

unsigned KCanFrame::Value() const
{
  if (Len < 2 || Len > 7)
    return 0;

  if (Len < 7)
    return 256*Data[Len-2] + Data[Len-1];

  if (Data[2] == 0xfa)
    return 256*Data[5] + Data[6];
  else
    return 256*Data[3] + Data[4];
}

int KCanFrame::Index() const
{
  if (Len < 2 || Len > 7 || Len < 5)
    return -1;

  if (Data[2] == 0xfa)
    return 256*Data[3] + Data[4];
  else
    return Data[2];
}

unsigned short KCanFrame::ReceiverId() const
{
  return 8*(Data[0] & 0xf0) +
         (Data[1] & 0x0f);
}

bool KCanFrame::GetDataFromStream(const char * str)
{
  Len = 0;
  for ( ; Len < 8; Len++)
  {
    while (*str == ' ' || *str == '.')
      str++;
    
    if (!NCanUtils::GetHexByte(str, Data[Len]))
      break;
  }
  while (*str == ' ' || *str == '.')
    str++;
  
  return !*str || *str == '\n' || *str == '\r';
}

void KCanFrame::FormMySqlInsert(char * statement) const
{
  char fields[256];
  char values[256];
  char str[64];
  
  NCanUtils::PrintMs(TimeStampDay, TimeStampMs, str);
  strcpy(fields, "counter,timestamp,can_id");
  
  sprintf(values, "%u,%s,%u",
          Counter & ~ 0x80000000, str, Id & ~ 0x80000000);
  for (int i = 0; i < Len; i++)
  {
    sprintf(fields + strlen(fields), ",d%d", i);
    sprintf(values + strlen(values), ",%d", Data[i]);
  }
  sprintf(statement, "INSERT can_log(%s) VALUES(%s);", fields, values);
}

void KCanFrame::FormHttpVariables(char * stream) const
{
  char time_str[64];
  
  NCanUtils::PrintMs(TimeStampDay, TimeStampMs, time_str);
  sprintf(stream, "counter=%d&timestamp=%s&can_id=%d",
          Counter & ~ 0x80000000, time_str, Id & ~ 0x80000000);
  
  for (int i = 0; i < Len; i++)
  {
    sprintf(stream + strlen(stream), "&d%d=%d", i, Data[i]);
  }
}


////////////////////////////////////////////////////////////////////////////////

static KComfortFrame InitFrame = {
  {0x0d, 0x00, 0x0d, 0x01, 0x00, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x26}
};

unsigned short KComfortFrame::InitId = 7077; // 0x089c

unsigned KComfortFrame::GetCheckSum() const
{
  return 256*Data[10] + Data[11];
}

unsigned KComfortFrame::GetSum() const
{
  unsigned Sum = 0;
  for (int i = 0; i < 10; i++)
    Sum += Data[i];
  
  return Sum;
}

bool KComfortFrame::CheckSum() const
{
  return GetSum() == GetCheckSum();
}

void KComfortFrame::SetCheckSum()
{
  unsigned Sum = GetSum();
  
  Data[10] = (unsigned char)(Sum >> 8);
  Data[11] = (unsigned char)(Sum & 0xff);
}

bool KComfortFrame::SetCanFrame(KCanFrame & Frame) const
{
  Frame.Init();
  Frame.Len = 7;
  
  Frame.Id = ((Data[0] & 0x0f) << 7) + (Data[1] & 0xf);
  Frame.Data[0] = (Data[2] << 4) + (Data[3] & 0x0f);
  
  for (int i = 0; i < 6; i++)
    Frame.Data[i + 1] = Data[i + 4];
  
  return CheckSum() &&
         (Data[0] & 0xf0) == 0;  // 11-Bit-Id
}

bool KComfortFrame::GetCanFrame(const KCanFrame & Frame)
{
  Data[0] = (Frame.Id >> 7) & 0x0f;
  Data[1] = Frame.Id & 0x0f;
  Data[2] = Frame.Data[0] >> 4;  
  Data[3] = Frame.Data[0] & 0x0f; // Sollte zum Abfragen 1 und für eine Antwort 2 sein!
  
  for (int i = 0; i < 6; i++)
    Data[i + 4] = Frame.Data[i + 1];
  
  if (Frame.Len == 5)
  {
    Data[8] = 0x80; // Stopf-Bytes initialisieren (siehe Marc2012)
    Data[9] = 0x08;
  }
  
  SetCheckSum();
  
  return (Frame.Len == 7 || Frame.Len == 5) &&
         (Frame.Id & ~0x7ff) == 0;  // 11-Bit-Id
}

void KComfortFrame::SetOk(bool Ok)
{
  for (int i = 0; i < 10; i++)
    Data[i] = Ok ? 0x55 : 0xAA;
  
  SetCheckSum();  
}

void KComfortFrame::SetOkValue(unsigned Value)
{
  SetOk();
  
  Data[8] = (unsigned char)(Value & 0xff);
  Data[9] = (unsigned char)(Value >> 8);
  
  SetCheckSum();
}

bool KComfortFrame::EqualData(const KComfortFrame & Frame) const
{
  for (int i = 0; i < 12; i++)
    if (Data[i] != Frame.Data[i])
      return false;
  
  return true;
}

void KComfortFrame::SetToString(char * Str) const
{
  Str[0] = 0;
  for (int i = 0; i < 12; i++)
    sprintf(Str + strlen(Str), "%2.2x", Data[i]);
}

bool KComfortFrame::SetFromString(const char * Str, bool AppendCheck)
{
  for (int i = 0; i < 12; i++)
  {
    while (*Str == ' ' || *Str == '.') Str++;
    if (!NCanUtils::GetHexByte(Str, Data[i]))
    {
      if (!AppendCheck ||
          i != 10)
        return false;

      SetCheckSum();
      return true;
    }
  }
  return true;
}

bool KComfortFrame::IsInitFrame() const
{
  return EqualData(InitFrame);
}

void KComfortFrame::SetInitOk()
{
  SetOkValue(InitId);
}


KCan232Frame::KCan232Frame(const char * str)
{
  *this = str;
}

bool KCan232Frame::SetCanFrame(KCanFrame & Frame) const
{
  Frame.Init();
  
  const char * Ptr = msg + 1;
  unsigned char b[4];
  
  if (!NCanUtils::GetHexByte(Ptr, b[0]))
    return false;
  
  switch (msg[0])
  {
    case 't':
    case 'r':
      if (b[0] > 0x7f ||
          !NCanUtils::IsHexDigit(*Ptr))
        return false;
    
      Frame.Id = 16*b[0] + NCanUtils::GetHexDigit(*Ptr++);
      break;
      
    case 'T':
    case 'R':
      if (b[0] > 0x3f ||
          !NCanUtils::GetHexByte(Ptr, b[1]) ||
          !NCanUtils::GetHexByte(Ptr, b[2]) ||
          !NCanUtils::GetHexByte(Ptr, b[3]))
        return false;
    
      for (int i = 0; i < 4; i++)
        Frame.Id = 256*Frame.Id + b[i];
      Frame.Id |= 0x80000000;
      break;
      
    default:
      return false;
  }
  if (msg[0] == 'r' || msg[0] == 'R') // RTR
    Frame.Id |= 0x40000000;
  
  if (*Ptr < '0' || *Ptr > '8')
    return false;
  
  Frame.Len = *Ptr++ - '0';
  
  for (int i = 0; i < Frame.Len; i++)
    if (!NCanUtils::GetHexByte(Ptr, Frame.Data[i]))
      return false;
  
  return true;  
}

bool KCan232Frame::IsOk() const
{
  KCanFrame Frame;
  
  return SetCanFrame(Frame);
}

bool KCan232Frame::GetCanFrame(const KCanFrame & Frame)
{
  if (Frame.Id & 0x80000000)
    sprintf(msg, "T%8.8X%1.1X", Frame.Id & 0x3ffffff, Frame.Len & 0xf);
  else
    sprintf(msg, "t%3.3X%1.1X", Frame.Id & 0x7ff, Frame.Len & 0xf);
 
  // RTR
  if (Frame.Id & 0x40000000)
    msg[0] += 'R' - 'T';
  
  for (int i = 0; i < Frame.Len; i++)
    sprintf(msg + strlen(msg), "%2.2X", Frame.Data[i]);
  
  return true;
}

KCan232Frame & KCan232Frame::operator =(const char * str)
{
  unsigned l = 0;
  int m = 0;
  while (str && str[m] && l < High(msg))
  {
    if (str[m] > ' ')
      msg[l++] = str[m];
    m++;
  }
  msg[l] = 0;
  
  return *this;
}

#if (0)

// Test zu "timestamp" des Volkszählers
int main(int argc, char* argv[])
{
  int Days;
  
  // 1334416684000.  14. April 2012  17:19
  //
  // ./ms2time.sh 1399346245967  ==> 2014-05-06 03:17:25
  //              1399349845000
  //                   -36
  //              1399353445000
  //              1399334400000
  //                   11800
  // date -d "2014-05-06 05:17:25" ==>  1399346245
  
  if (NCanUtils::GetArtDays(2012, 4, 14, Days))
  {
    char str[64];
    char mstr[64];
    
    int sec = 60*(60*15 + 18) + 4;
    
    NCanUtils::PrintTime((int) Days, 1000*sec, str, false);
    NCanUtils::PrintMs((int) Days, 1000*sec, mstr);
    printf("%d   %s   %s\n", sec, mstr, str);
  }
  
  return 0;
}

#endif
